/*******************************************************************************
 * Copyright 2005, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     The Chisel Group, University of Victoria
 *     Prefuse: Jeffrey Heer, Univeristy of California & Berkeley 
 *     			http://prefuse.sourceforge.net/index.html
 *******************************************************************************/
package org.eclipse.mylar.zest.layouts.algorithms;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.StringWriter;

import org.eclipse.mylar.zest.layouts.LayoutStyles;
import org.eclipse.mylar.zest.layouts.dataStructures.DisplayIndependentRectangle;
import org.eclipse.mylar.zest.layouts.dataStructures.InternalNode;
import org.eclipse.mylar.zest.layouts.dataStructures.InternalRelationship;

import edu.berkeley.guir.prefusex.force.DragForce;
import edu.berkeley.guir.prefusex.force.ForceItem;
import edu.berkeley.guir.prefusex.force.ForceSimulator;
import edu.berkeley.guir.prefusex.force.NBodyForce;
import edu.berkeley.guir.prefusex.force.SpringForce;

/**
 * This is a modified version of the Prefuse ForceDirectedLayout class. 
 * It was rewritten to remove the java.awt dependency.
 * 
 * @author irbull 
 * @author ccallendar
 */
public class ForceDirectedAlgorithm extends ContinuousLayoutAlgorithm {

	/** The default minimum spring length (30). */
	public static final float DEFAULT_MIN_SPRING_LENGTH = 30F;
	/** The default maximum spring length (300). */
	public static final float DEFAULT_MAX_SPRING_LENGTH = 300F;
	/** 
	 * The default minimum change in position of a node.  This is used to cause 
	 * the algorithm to stabilize (the nodes stop moving when their change in
	 * in position is smaller than this distance).
	 */
	public static final double DEFAULT_MIN_STABILIZE_DISTANCE = 0.2D; 

	private static final String FORCE_ITEM = "forceItem";
	
	private boolean enforceBounds;
    private ForceSimulator forceSim;
	private int runCount;
	private int maxRuns;
	private boolean startInCenter;
	private float minimumSpringLength;
	private float maximumSpringLength;
	private double minimumStabilizeDistance;
	private boolean overlapAllowed;
	private boolean stabilize;
	
    private DisplayIndependentRectangle layoutBounds = null;

    public ForceDirectedAlgorithm() {
		this(LayoutStyles.NONE);
	}
    
    public ForceDirectedAlgorithm(int styles) {
		this(styles, true, 0);
	}
	
	/**
	 * Initializes the ForceDirectedAlgorithm object.  Adds NBodyForce, SpringForce and DragForce to
	 * the ForceSimulator.  Sets the minimum and maximum spring length to be the default values.
	 * @param enforceBounds	If the bounds should be enforced.
	 * @param maxRuns		The max number of times to run the algorithm.  
	 */
	public ForceDirectedAlgorithm(int styles, boolean enforceBounds, int maxRuns) {
		super( styles );
		this.enforceBounds = enforceBounds;
		setMaxRuns(maxRuns);
		this.runCount = 0;
		this.startInCenter = false;
		this.minimumSpringLength = DEFAULT_MIN_SPRING_LENGTH;
		this.maximumSpringLength = DEFAULT_MAX_SPRING_LENGTH;
		this.minimumStabilizeDistance = DEFAULT_MIN_STABILIZE_DISTANCE;
		this.overlapAllowed = true;
		this.stabilize = false;
		this.forceSim = new ForceSimulator();
		forceSim.addForce(new NBodyForce());
		forceSim.addForce(new SpringForce());
		forceSim.addForce(new DragForce());
	}
	
	/**
	 * Updates the bounds within which the algorithm runs.
	 */
	public void setLayoutArea(double x, double y, double width, double height) {
		layoutBounds = new DisplayIndependentRectangle(x, y, width, height);
		//TODO: should all the nodes be moved to fill the area?
	}
	
	protected void preLayoutAlgorithm(InternalNode[] entitiesToLayout, InternalRelationship[] relationshipsToConsider, double x, double y, double width, double height) {
        
        // the layout might have already been set
        if (layoutBounds == null) {
        	layoutBounds = new DisplayIndependentRectangle(x, y, width, height);
        }
        
		// set all the nodes to start at the center of the bounds
		if (startInCenter) {
			double centerX = x + (width/2);
			double centerY = y + (height/2);
            for (int i = 0; i < entitiesToLayout.length; i++) {
                InternalNode node = entitiesToLayout[i];
	            node.setInternalLocation(centerX, centerY);
	        }
		}
		
		this.runCount = 0;
	}

	protected void computeOneIteration(InternalNode[] entitiesToLayout, InternalRelationship[] relationshipsToConsider, double x, double y, double width, double height) {
		
		sleepWhilePaused();
		
		// update the run counter
		runCount++;
		
		try {
			forceSim.clear();
			initSimulator(entitiesToLayout, relationshipsToConsider, forceSim);
			forceSim.runSimulator(50);
		} catch (StackOverflowError sofe) {
			// sometimes get a StackOverflowError in NBodyForce insertHelper() or insert() methods 
			System.err.println("StackOverflowError!");
			StringWriter stringWriter = new StringWriter();
			sofe.printStackTrace(new PrintWriter(stringWriter));
			outputNodesToFile(entitiesToLayout, stringWriter.toString().substring(0, 600));
			return;
		}

		// move the entities to the new positions given inside the ForceItems
        updateNodePositions(entitiesToLayout);

		
		// Let the other thread get some CPU time otherwise there is a context switch occasionally
		try {
			//Thread.yield();
			Thread.sleep(50);
		} catch (Exception e) {
		}		
	}
	
	protected int getCurrentLayoutStep() {
		return runCount;
	}
	
	protected int getTotalNumberOfLayoutSteps() {
		return maxRuns;
	}
	
	protected boolean performAnotherNonContinuousIteration() {
		return ((maxRuns == -1) || (runCount < maxRuns));
	}
	
	protected void postLayoutAlgorithm(InternalNode[] entitiesToLayout, InternalRelationship[] relationshipsToConsider) {
		// NOOP
	}
	  
	/**
	 * Iterates through all the nodes updating their positions with the
	 * location given inside the ForceItems.
	 */
	protected void updateNodePositions(InternalNode[] entities) {
        double x1 = layoutBounds.x; 
		double y1 = layoutBounds.y;
        double x2 = layoutBounds.x + layoutBounds.width; 
		double y2 = layoutBounds.y + layoutBounds.height;
        
        // update node positions from the ForceItems
        for (int i = 0; i < entities.length; i++) {
            InternalNode node = entities[i];
            ForceItem fitem = (ForceItem)node.getAttributeInLayout(FORCE_ITEM);
            double x = fitem.location[0];
            double y = fitem.location[1];
            if (enforceBounds) {
                if (x > x2) { 
					x = x2;
                } else if (x < x1) {
					x = x1;
                }
                if (y > y2) { 
					y = y2;
                } else if (y < y1) { 
					y = y1;
                }
            }
            
            if (stabilize) {
                // check the distance between the new point and the old point
            	// if it is smaller than the minimum difference then don't update the position
            	// this will cause the algorithm to "stabilize" when the nodes aren't moving much
                double dr = Math.sqrt(Math.pow(node.getInternalX()-x, 2) + Math.pow(node.getInternalY()-y, 2));
                if (dr > minimumStabilizeDistance) {
                   	node.setInternalLocation( x, y );
                } else {
                }
            } else {
            	node.setInternalLocation( x, y );
            }
        }
    }

	/**
	 * Writes out the nodes positions to a file called stackoverflow.txt stored in
	 * the temporary directory.
	 */
	protected synchronized void outputNodesToFile(InternalNode[] entities, String extra) {
		BufferedWriter out = null;
		try {
			File temp = File.createTempFile("ForceDirectedAlgorithm", "");
			File file = new File(temp.getParentFile(), "forcedirected_output.txt");
			System.out.println("Outputting node positions to: '" + file.getAbsolutePath() + "'");
			out = new BufferedWriter(new FileWriter(file, true));
			out.write("=================================================");
	        out.newLine();
			out.write((new java.util.Date()).toString());
	        out.newLine();
			out.write("=================================================");
	        out.newLine();
	        if ((extra != null) && (extra.length() > 0)) {
	        	out.write(extra);
	        	out.newLine();
				out.write("=================================================");
	        	out.newLine();
	        }
	        // update node positions from the ForceItems
            for (int i = 0; i < entities.length; i++) {
                InternalNode node = entities[i];
	            ForceItem fitem = (ForceItem)node.getAttributeInLayout(FORCE_ITEM);
	            double x = fitem.location[0];
	            double y = fitem.location[1];
	            String s = node.toString();
	            out.write(s);
	            for (int j = s.length(); j < 30; j++) {
	            	out.write(" ");
	            }
	            out.write(x + "," + y);
		        out.newLine();
	        }
	        out.newLine();
	        out.close();
	        temp.delete();
		} catch (Exception ex) {
			ex.printStackTrace();
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (Exception ex) {}
			}
		}
	}
	    
    /**
     * Loads the simulator with all relevant force items and springs.
     * The ForceItems are stored inside the nodes using the setAttributeInLayout() method.
     */
    protected void initSimulator(InternalNode[] entities, InternalRelationship [] relationships, ForceSimulator fsim) {
        for (int i = 0; i < entities.length; i++) {
           InternalNode node = entities[i];
           ForceItem fitem = (ForceItem) node.getAttributeInLayout(FORCE_ITEM);
           if (fitem == null) {
               fitem = new ForceItem();			   
               node.setAttributeInLayout(FORCE_ITEM, fitem);
           }
           double x = node.getInternalX();
           double y = node.getInternalY();
		   if (node.hasPreferredLocation()) {
			   x = node.getPreferredX();
			   y = node.getPreferredY();
		   }
           fitem.location[0] = (Double.isNaN(x) ? 0f : (float)x);
           fitem.location[1] = (Double.isNaN(y) ? 0f : (float)y);
           fsim.addItem(fitem);
       }
       for (int i = 0; i < relationships.length; i++) {
           InternalRelationship edge = relationships[i];
		   InternalNode n1 = edge.getSource();
           ForceItem f1 = (ForceItem) n1.getAttributeInLayout(FORCE_ITEM);
		   InternalNode n2 = edge.getDestination();
           ForceItem f2 = (ForceItem) n2.getAttributeInLayout(FORCE_ITEM);
           
           // get the spring length add add the spring if the length isn't -1
           float springLength = getSpringLength(n1, n2, edge);
           if (springLength != -1F) {
        	   fsim.addSpring(f1, f2, springLength);
           }
       }
       
       // if overlapping nodes aren't allowed, then put an "invisible" spring between
       // any nodes who are too close together.  This doesn't ensure that ndoes won't overlap
       // but will try and keep nodes as far apart as they can go
       if (!overlapAllowed) {
	       for (int i = 0; i < entities.length; i++) {
	    	   for (int j = i+1; j < entities.length; j++) {
	    		   InternalNode n1 = entities[i];
	    		   InternalNode n2 = entities[j];
	    		   int min = (int) getMinDistance(n1, n2);
	    		   int dist = (int) getDistance(n1, n2);
	    		   if (dist < min) {
	    			   //System.out.println(dist + " < " + min + "  " + n1 + " <-> " + n2);
	    			   ForceItem f1 = (ForceItem) n1.getAttributeInLayout(FORCE_ITEM);
	    			   ForceItem f2 = (ForceItem) n2.getAttributeInLayout(FORCE_ITEM);
	    			   fsim.addSpring(f1, f2, 1E-4f, (float)min);
	    		   }
	    	   }
	       }
       }
    }
    
	/**
	 * Gets the spring length.  The spring length will be between the minimum and maximum lengths depending on the 
	 * value of the edge weight.  If the edge weight is -1 then the spring length is -1 meaning no spring tension.
	 * @param n1		The first node.
	 * @param n2		The second node.
	 * @param edge		The relationship between the nodes.
	 * @return The spring length.
	 */
    protected float getSpringLength(InternalNode n1, InternalNode n2, InternalRelationship edge) {
		float len = -1F;
		double weight = edge.getWeight();		// {-1, [0-1]}
		if ((weight >= 0) && (weight <= 1)) {
			float diff = maximumSpringLength - minimumSpringLength;
			len = maximumSpringLength - (float)(diff * edge.getWeight());
		}
        return len;
    }
    
    /**
     * Gets the minimum distance between the two nodes.  This distance
     * is the distance from two centers if they were touching only at the corner
     * of their bounds.
     * @param n1
     * @param n2
     * @return the minimum distance to ensure no overlap
     */
    protected double getMinDistance(InternalNode n1, InternalNode n2) {
		double w1 = n1.getInternalWidth() / 2;
		double h1 = n1.getInternalHeight() / 2;
		double w2 = n2.getInternalWidth() / 2;
		double h2 = n2.getInternalHeight() / 2;
		double d1 = Math.sqrt(Math.pow(w1, 2) + Math.pow(h1, 2));
		double d2 = Math.sqrt(Math.pow(w2, 2) + Math.pow(h2, 2));
		return (d1 + d2);
    }
    
    /**
     * Gets the actual distance between the two nodes.
     * @param n1
     * @param n2
     * @return the distance between the nodes.
     */
    protected double getDistance(InternalNode n1, InternalNode n2) {
		double x1 = n1.getInternalX();
		double y1 = n1.getInternalY();
		double x2 = n2.getInternalX();
		double y2 = n2.getInternalY();
    	return Math.sqrt(Math.pow(x2-x1, 2) + Math.pow(y2-y1, 2));
    }

	/**
	 * Sets the force simulator.
	 * @param forceSim
	 */
	public void setForceSimulator(ForceSimulator forceSim) {
		this.forceSim = forceSim;
	}
	
	/**
	 * Returns the current force simulator.
	 * @return ForceSimulator
	 */
	public ForceSimulator getForceSimulator() {
		return forceSim;
	}

	/**
	 * Returns the number of times this algorithm is set to run.
	 * @return int
	 */
	public int getMaxRuns() {
		return maxRuns;
	}

	/**
	 * Sets the run count.
	 * @param maxRuns
	 */
	public void setMaxRuns(int maxRuns) {
		if ( maxRuns < 0 ) throw new RuntimeException("Maxruns must be >= 0");
		this.maxRuns = (maxRuns > 0 ? maxRuns : -1);
	}

	/**
	 *  Returns if the nodes should start in the center of the bounds
	 * @return boolean
	 */
	public boolean isStartInCenter() {
		return startInCenter;
	}

	/**
	 * Sets if the nodes should initially start in the center of the bounds.
	 * @param startInCenter
	 */
	public void setStartInCenter(boolean startInCenter) {
		this.startInCenter = startInCenter;
	}

	/**
	 * Returns the minimum spring length.
	 * @return minimum spring length.
	 */
	public float getMinimumSpringLength() {
		return minimumSpringLength;
	}

	/**
	 * Returns the maximum spring length.
	 * @return maximum spring length.
	 */
	public float getMaximumSpringLength() {
		return maximumSpringLength;
	}
	
	/**
	 * Sets the minimum and maximum spring lengths.
	 * @param minSpringLength	The min spring length (must be >= 1)
	 * @param maxSpringLength	The max spring length (must be >= minSpringLength)
	 */
	public void setSpringLengthRange(float minSpringLength, float maxSpringLength) {
		this.minimumSpringLength = (minSpringLength >= 1 ? minSpringLength : 1);
		this.maximumSpringLength = (maxSpringLength >= this.minimumSpringLength ? maxSpringLength : this.minimumSpringLength);
	}
	
	public double getMinimumStabilizeDistance() {
		return minimumStabilizeDistance;
	}
	
	/**
	 * Sets the minimum change in position of a node when the nodes will
	 * cease moving.  This only has an affect if stabilize is true.
	 * @see #setStabilize(boolean)
	 * @see #DEFAULT_MIN_STABILIZE_DISTANCE
	 * @param minDistance
	 */
	public void setMinimumStabilizeDistance(double minDistance) {
		this.minimumStabilizeDistance = minDistance;
	}
	
	/**
	 * Returns if the algorithm is set to stabilize.  
	 * This means that nodes whose position change is less than the
	 * minimum distance won't move.
	 * @see #setMinimumStabilizeDistance(double)
	 * @return boolean if the algorithm is set to stabilize
	 */
	public boolean isStabilize() {
		return stabilize;
	}
	
	/**
	 * Sets if the algorithm should stabilize.  
	 * If stabilize is true it means that nodes whose position change 
	 * is less than the minimum distance won't move.
	 * @see #setMinimumStabilizeDistance(double)
	 * @param stabilize
	 */
	public void setStabilize(boolean stabilize) {
		this.stabilize = stabilize;
	}
	
	/**
	 * Returns if nodes are allowed to overlap.
	 * @return boolean
	 */
	public boolean isOverlappingNodesAllowed() {
		return overlapAllowed;
	}
	
	/**
	 * Sets if nodes are allowed to overlap.
	 * This doesn't ensure that nodes won't overlap but aims
	 * to prevent it.
	 * @param allowed
	 */
	public void setOverlappingNodesAllowed(boolean allowed) {
		this.overlapAllowed = allowed;
	}
	
	public boolean isEnforceBounds() {
		return enforceBounds;
	}

	/**
	 * Sets whether to enforce the bounds.  Any nodes that move outside the bounds
	 * are moved back inside the bounds.
	 * @param enforceBounds
	 */
	public void setEnforceBounds(boolean enforceBounds) {
		this.enforceBounds = enforceBounds;
	}
	
	protected boolean isValidConfiguration(boolean asynchronous, boolean continuous) {
		if ( asynchronous && continuous ) return true;
		else if ( asynchronous && !continuous ) return true;
		else if ( !asynchronous && continuous ) return false;
		else if ( !asynchronous && !continuous ) return true;
		
		return false;
	}

	
}
